/* $Header: /fridge/cvs/cgi.targeted/targeted-cgi.c,v 1.6 2001/12/10 23:08:41 justins Exp $ */
/*
   
   targeted-cgi.c             Copyright(c) 2001-1998 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Utilities for manipulating the CGI arguments
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#include <assert.h>
#include <libjtree.h>
#include <targeted-config.h>


const char *lookup_cgi_argument(const bintree *t, const char *arg, const char *def) {
/* Lookup a CGI argument from the query string; if it 
   is not found, then return the default value given. */
   
   const bintreenode *node;
   
   assert(t != NULL);
   assert(arg != NULL);
   assert(def != NULL);
   
   /* Lookup the requested argument */
   node = bintree_pair_find_first(t, arg);
   if(node == NULL) return(def);
   return(((bintreepair *)node->data)->value);
   
}


void set_cgi_argument(bintree *t, const char *arg, const char *val) {
/* Sets a CGI argument to the specified value. */

   const bintreenode *node;
   
   assert(t != NULL);
   assert(arg != NULL);
   assert(val != NULL);
   
   /* Set the argument to the specified value */
   node = bintree_pair_find_first(t, arg);
   if(node == NULL) {
      bintree_insert(t, pair_alloc(arg, val));
   } else {
      pair_update_value((bintreepair *)node->data, val);
   }
   
}


void set_cgi_argument_int(bintree *t, const char *arg, int val) {
/* Same as above, but initial argument is an integer */

   char buf[0x1000];
   sbprintf(buf, sizeof(buf), "%d", val);
   set_cgi_argument(t, arg, buf);

}


void unset_cgi_argument(bintree *t, const char *key) {
/* Clear the CGI argument specified */

   assert(t != NULL);
   assert(key != NULL);
   bintree_pair_delete(t, key, true);

}


typedef struct _cgi_url_data {
   char *buf;
   int size;
   int curlen;
   int count;
   bool noncgi;
} cgi_url_data;


static bool _construct_cgi_url(const void *_data, void *_params) {

   const bintreepair *data = (const bintreepair *)_data;
   cgi_url_data *params = (cgi_url_data *)_params;

   if(!params->noncgi || !(strequal(data->key, "mode") || strequal(data->key, "path"))) {
      sbprintf(params->buf + params->curlen, params->size - params->curlen, "%s%s=%s", params->count == 0 ? "?" : "&amp;", data->key, data->value);
      params->curlen += strlenn(params->buf + params->curlen);
      ++params->count;
   }
   return(true);

}


void construct_cgi_url(const config *c, const bintree *t, char *buf, int size) {
/* Form a CGI URL with the arguments currently in the query_args. */

   cgi_url_data params;
   const char *mode;
   const char *path;
   bool using_noncgi_form = false;
   
   assert(c != NULL);
   assert(buf != NULL);
   assert(size > 0);

   /* Copy the scriptname to the URL */
   *buf = '\0';
   if(c->use_noncgi_url) {
      /* Rewrite basename as a path if applicable */
      mode = lookup_cgi_argument(t, "mode", "");
      path = lookup_cgi_argument(t, "path", "/");
      if(strequal(mode, "viewimage")) {
         sbprintf(buf, size, "/targeted-image%s", path);
         using_noncgi_form = true;
      } else if(strequal(mode, "thumb")) {
         sbprintf(buf, size, "/targeted-thumb%s", path);
         using_noncgi_form = true;
      } else if(strequal(mode, "view")) {
         sbprintf(buf, size, "/targeted-view%s", path);
         using_noncgi_form = true;
      }
   }

   if(!using_noncgi_form) {
      strcopyb(buf, stra(c->cgi_uri), size);
   }
   
   /* Append the arguments to the URL */
   params.buf = buf;
   params.size = size;
   params.curlen = strlenn(buf);
   params.count = 0;
   params.noncgi = using_noncgi_form;
   bintree_walk(t, _construct_cgi_url, &params);

}


bintree *new_cgi_arguments() {

   return(bintree_new(bintree_pair_lessthan, (destructorfn)pair_free));

}


static bool _add_cgi_pair(const void *_data, void *_param) {

   const bintreepair *data = (const bintreepair *)_data;
   bintree *t = (bintree *)_param;
   
   bintree_insert(t, pair_alloc(data->key, data->value));
   return(true);

}


bintree *clone_cgi_arguments(const bintree *t) {
/* Clone the CGI arguments stored in the paired bintree indicated */

   bintree *new;

   assert(t != NULL);
   
   new = new_cgi_arguments();
   if(new == NULL) return(NULL);
   
   bintree_walk(t, _add_cgi_pair, new);
   unset_cgi_argument(new, "view");
   unset_cgi_argument(new, "thumb");
   unset_cgi_argument(new, "viewimage");
   unset_cgi_argument(new, "tree");
   unset_cgi_argument(new, "depth");
   unset_cgi_argument(new, "width");
   unset_cgi_argument(new, "height");
   return(new);
   
}
