/* $Header: /fridge/cvs/cgi.targeted/targeted-image.c,v 1.10 2001/12/10 23:08:41 justins Exp $ */
/*
   
   targeted-image.c           Copyright(c) 2001-1998 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Routines to lookup useful facts about images, and do image caching


   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/* Includes */
#include <targeted-image.h>
#include <targeted-util.h>
#include <targeted-view.h>
#include <targeted-log.h>
#include <targeted-cgi.h>
#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <magick/api.h>
#include <errno.h>


/* Version number for index files */
#define  DATA_VERSION   2


/* Default quality for JPEG thumbnail images */
#define  THUMB_QUALITY  60


/* Quality and maximum sizes for dynamically resized images */
#define  RESIZE_QUALITY 85
#define  RESIZE_MAX_W   8192
#define  RESIZE_MAX_H   8192


static void _index_base_file_name(const config *c, char *tfile, int size, const char *r_path) {
/* Compute the basename of the index filename based on the real 
   name given.  The given name should be relative, and the 
   resulting name should be prepended by either the real CGI
   directory name, or the web-visible name.  */

   char path[0x1000];
   char *p;

   /* Replace all path separators with colons */
   strcopyb(path, r_path, sizeof(path));
   p = path;
   while(*p != '\0') {
      if(*p == '/') *p = ':';
      ++p;
   }
   
   /* Construct the final name */
   sbprintf(tfile, size, "image.%s", path);

}


static void _index_file_name(const config *c, char *tfile, int size, const char *r_path) {
/* Compute the index filename based on the real name given.  The
   given name should be relative, but the resulting name will be
   a real filesystem name.  */

   char buf[0x1000];
   _index_base_file_name(c, buf, sizeof(buf), r_path);
   sbprintf(tfile, size, "%s/%s::data", stra(c->cache_fs), buf);

}


static void _index_file_name_using_path(const config *c, char *tfile, int size, const char *path, const char *r_path) {
/* See next two functions */

   char buf[0x1000];
   const char *p;
   _index_base_file_name(c, buf, sizeof(buf), r_path);

   /* Check the extension and create a matching extn */
   if(strlenn(r_path) < 4) {
      sbprintf(tfile, size, "%s/%s::thumb.gif", path, buf);
   } else {
      p = r_path + strlenn(r_path) - 4;
      if(strequal(p, ".jpg")) {
         sbprintf(tfile, size, "%s/%s::thumb.jpg", path, buf);
      } else if(strequal(p, ".png")) {
         sbprintf(tfile, size, "%s/%s::thumb.png", path, buf);
      } else if(strequal(p, ".gif")) {
         sbprintf(tfile, size, "%s/%s::thumb.gif", path, buf);
      } else {
         sbprintf(tfile, size, "%s/%s::thumb.png", path, buf);
      }
   } /* Image extension exists? */

}


static void _index_real_thumb_file_name(const config *c, char *tfile, int size, const char *r_path) {
/* Compute the image filename based on the real name given.  The
   given name should be relative, but the resulting name will be
   a real filesystem name.  */

   _index_file_name_using_path(c, tfile, size, stra(c->cache_fs), r_path);

}


static void _index_thumb_file_name(const config *c, char *tfile, int size, const char *r_path) {
/* Compute the web-visible thumbnail filename.  */

   _index_file_name_using_path(c, tfile, size, stra(c->cache_uri), r_path);

}


static bool _get_real_image_size(const config *c, const char *tfile, const char *file, const char *r_path, image_desc *desc) {
/* Collect data from the real image */

   char buf[0x1000];
   char tfilename[0x1000];
   ExceptionInfo exception;
   ImageInfo *image_info;
   Image *image;
   Image *timage;
   struct stat s;
   FILE *f;
   
   /* Initialize the image info structure and read an image. */
   GetExceptionInfo(&exception);
   image_info = CloneImageInfo(NULL);
   strcopyb(image_info->filename, file, MaxTextExtent);
   image = ReadImage(image_info, &exception);
   if(image == NULL) return(false);

   /* Get the image dimensions */
   desc->width  = image->magick_columns;
   desc->height = image->magick_rows;
   if(desc->width <= c->thumbnail_width) {
      /* No thumbnail necessary */
      desc->thumbwidth  = desc->width;
      desc->thumbheight = desc->height;
      strcopya(desc->thumb, "");
      strcopyb(tfilename, "", sizeof(tfile));
   } else {
      /* We will try to write a thumbnail */
      desc->thumbwidth  = c->thumbnail_width;
      desc->thumbheight = desc->height * desc->thumbwidth / desc->width;
      _index_real_thumb_file_name(c, tfilename, sizeof(tfilename), r_path);
   
      /* Attempt to scale and save a thumbnail */
      timage = ResizeImage(image, desc->thumbwidth, desc->thumbheight, LanczosFilter, 1.0, &exception);
      if(timage != NULL) {
         strcopyb(image_info->filename, tfilename, MaxTextExtent);
         strcopyb(timage->filename, tfilename, MaxTextExtent);
         image_info->quality = THUMB_QUALITY;
         if(WriteImage(image_info, timage) == 0) {
            targeted_log_error(c, "unable to write thumb file for image data");
            strcopya(desc->thumb, "");
         } else {
            targeted_log_message(c, "wrote thumb file for image data");
            _index_thumb_file_name(c, buf, sizeof(buf), r_path);
            strcopya(desc->thumb, buf);
         }
         DestroyImage(timage);
      }
   } /* Write a thumbnail? */
   
   /* Release the image from memory */
   DestroyImage(image);
   DestroyImageInfo(image_info);
   
   /* Get file size statistics */
   desc->size = 0;
   desc->thumbsize = 0;
   if(stat(file, &s) != 0) {
      targeted_log_error(c, "unable to collect statistics on image real file");
      return(true);
   }
   desc->size = s.st_size;
   if(*tfilename != '\0') {
      if(stat(tfilename, &s) != 0) {
         targeted_log_error(c, "unable to collect statistics on image thumbnail file");
         return(true);
      }
      desc->thumbsize = s.st_size;
   }

   /* Attempt to write an index file */
   if((f = fopen(tfile, "w")) == NULL) {
      targeted_log_error(c, "unable to write index file for image data");
      return(true);
   }
   fprintf(f, "%d|%s|%s|%d|%d|%d|%d|%d|%d|%d\n", DATA_VERSION, r_path, stra(desc->thumb), desc->width, desc->height, desc->thumbwidth, desc->thumbheight, desc->size, desc->thumbsize, THUMB_QUALITY);
   fclose(f);
   targeted_log_message(c, "wrote index file for image data");
   
   /* This is usually slow, so flush output */
   fflush(stdout);

   /* Success */
   return(true);

}


image_desc *new_image_desc(const config *c, const char *r_path) {
/* Determine the size of a specified image file (relative path).
   Use the index file if possible since that is significantly
   faster loading. */

   stringa *arg;
   image_desc *desc;
   char buf[0x1000];
   char file[0x1000];
   char tfile[0x1000];
   struct stat tstat;
   struct stat istat;
   FILE *f;

   assert(c != NULL);
   assert(r_path != NULL);
   
   /* Get image file stats */
   qualify_full_path(c, file, sizeof(file), r_path);
   if(stat(file, &istat) != 0) return(NULL);
   
   /* allocate descriptive structure */
   desc = (image_desc *)malloc(sizeof(image_desc));
   if(desc == NULL) return(NULL);
   desc->thumb = strnewa(NULL);

   /* Get stats on the targeted index (if exists) */
   _index_file_name(c, tfile, sizeof(tfile), r_path);
   if(stat(tfile, &tstat) == 0) {
      /* Targeted index exists; is it newer? */
      if(tstat.st_mtime > istat.st_mtime) {
         /* Index file is newer; let's use it! */
         if((f = fopen(tfile, "r")) != NULL) {
            /* Read the index file data */
            fgets(buf, sizeof(buf), f);
            fclose(f);
            if(getnumargs(buf, '|') >= 10) {
               /* Load data into desc */
               arg = strnewa(NULL);
               getarga(arg, buf, 0, '|');
               /* Version number check */
               if(atoi(stra(arg)) == DATA_VERSION) {
                  getarga(desc->thumb, buf, 2, '|');
                  getarga(arg, buf, 3, '|');
                  desc->width = atoi(stra(arg));
                  getarga(arg, buf, 4, '|');
                  desc->height = atoi(stra(arg));
                  getarga(arg, buf, 5, '|');
                  desc->thumbwidth = atoi(stra(arg));
                  getarga(arg, buf, 6, '|');
                  desc->thumbheight = atoi(stra(arg));
                  getarga(arg, buf, 7, '|');
                  desc->size = atoi(stra(arg));
                  getarga(arg, buf, 8, '|');
                  desc->thumbsize = atoi(stra(arg));
                  strreleasea(&arg);
               
                  /* Validate the cache filename */
                  if(!strequal(stra(desc->thumb), "")) {
                     _index_thumb_file_name(c, file, sizeof(file), r_path);
                     strcopya(desc->thumb, file);
                  }
               
                  /* Success */
                  return(desc);
               } else {
                  /* Failure; bad version */
                  strreleasea(&arg);
               } /* Version check */
            } /* Bogus data format */
         } else {
            targeted_log_error(c, "unable to read image index file");
         } /* Cannot read index file */
      } else {
         targeted_log_message(c, "image index data is older than the image");
      } /* Index file older than image */
   } /* No index file to be found */
   
   /* Do this the old-fashioned way */
   if(!_get_real_image_size(c, tfile, file, r_path, desc)) {
      free_image_desc(&desc);
      return(NULL);
   }
   return(desc);

}


void free_image_desc(image_desc **desc) {
/* Release an image description */

   if(desc == NULL || *desc == NULL) return;
   strreleasea(&(*desc)->thumb);
   free(*desc);
   *desc = NULL;

}


void init_image_system() {
/* Perform one-time initialisation of the imaging system */

   MagickIncarnate("targeted");

}


static bool _targeted_image(const config *c, const char *r_path, int width, int height) {
/* Transmit a rescaled version of the image specified by path=.  This
   recognizes the width and height options ONLY; the viewimage primitive
   should do all size translations before invoking this CGI.  This CGI
   is not really intended to be called directly... */

   char file[0x1000];
   char buf[0x1000];
   ExceptionInfo exception;
   ImageInfo *image_info;
   Image *image;
   Image *timage;
   
   /* Validate the image name */
   if(!targeted_view_okay(c, r_path, false)) {
      targeted_log_error(c, "not permitted to view this image");
      return(false);
   }
         
   /* Qualify the real filename of the image */
   qualify_full_path(c, file, sizeof(file), r_path);
   
   /* Validate the image dimensions */
   if(width < 0) width = 0;
   if(height < 0) height = 0;
   if(width > RESIZE_MAX_W) width = RESIZE_MAX_W;
   if(height > RESIZE_MAX_H) height = RESIZE_MAX_H;
   
   /* Initialize the image info structure and read an image. */
   GetExceptionInfo(&exception);
   image_info = CloneImageInfo(NULL);
   strcopyb(image_info->filename, file, MaxTextExtent);
   image = ReadImage(image_info, &exception);
   if(image == NULL) {
      targeted_log_error(c, "Failed to open image file");
      printf("Content-type: text/plain\n\n"
             "*** Malformed image command, aborting ***\n\n"
             "HINT:  The script would not normally generate a URL that generated\n"
             "       this error.  If you obtained this error by following a link\n"
             "       then contact me ASAP.  More likely you are mucking with the\n"
             "       URL manually, in which case you should *NOT* modify the URL\n"
             "       manually -- the system is *NOT* designed for manual URL entry!\n");
      return(false);
   }

   /* Determine the image format */
   if(strequal(image->magick, "GIF")) {
      printf("Content-type: image/gif\n\n");
   } else if(strequal(image->magick, "JPEG")) {
      printf("Content-type: image/jpeg\n\n");
   } else if(strequal(image->magick, "PNG")) {
      printf("Content-type: image/png\n\n");
   } else {
      targeted_log_error(c, "Unknown image type/cannot determine MIME type");
      printf("Content-type: text/plain\n\n"
             "*** Malformed image command, aborting ***\n\n"
             "HINT:  The script would not normally generate a URL that generated\n"
             "       this error.  If you obtained this error by following a link\n"
             "       then contact me ASAP.  More likely you are mucking with the\n"
             "       URL manually, in which case you should *NOT* modify the URL\n"
             "       manually -- the system is *NOT* designed for manual URL entry!\n\n");
      return(false);
   }
   fflush(stdout);

   /* Currently disallow expansion */
   if(width > image->magick_columns) width = image->magick_columns;
   if(height > image->magick_rows) width = image->magick_rows;

   /* Attempt to scale and save the image */
   timage = ResizeImage(image, width, height, LanczosFilter, 1.0, &exception);
   if(timage == NULL) return(false);

   /* Close out stdout to make sure we can write to it later */
   image_info->file = stdout;
   timage->file = stdout;
   image_info->quality = RESIZE_QUALITY;
   if(WriteImage(image_info, timage) == 0) {
      sbprintf(buf, sizeof(buf), "unable to write rescaled image to channel:  %s (%s)", 
               timage->exception.reason, timage->exception.description);
      targeted_log_error(c, buf);
      printf("Content-type: text/plain\n\n"
             "*** I/O error on image command, aborting ***\n\n"
             "This should not have occurred, and indicates a fault in the system\n"
             "libraries.  Contact me ASAP so I can correct the problem.\n");
      return(false);
   }
   targeted_log_message(c, "sent rescaled image over channel");
   
   /* Release the images from memory */
   DestroyImage(timage);
   DestroyImage(image);
   DestroyImageInfo(image_info);
   
   /* This is usually slow, so flush output */
   fflush(stdout);

   /* Success */
   return(true);

}


void targeted_image(const config *c) {
/* Interface for image system */

   const char *path;
   int width;
   int height;

   assert(c != NULL);
   
   path = lookup_cgi_argument(c->env.query_args, "path", "");
   width = atoi(lookup_cgi_argument(c->env.query_args, "width", "0"));
   height = atoi(lookup_cgi_argument(c->env.query_args, "height", "0"));
   _targeted_image(c, path, width, height);

}
