/* $Header: /fridge/cvs/cgi.targeted/targeted-list.c,v 1.5 2001/12/10 23:08:41 justins Exp $ */
/*
   
   targeted-list.c            Copyright(c) 2001-1998 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Functions for managing directory lists
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/* Include files */
#include <targeted-list.h>
#include <targeted-util.h>
#include <targeted-log.h>
#include <sys/stat.h>
#include <unistd.h>
#include <dirent.h>


bool is_directory(const config *c, const char *r_path, const char *d_name) {
/* Returns true if the relative path listed is a directory, and
   the user has the appropriate access rights to it. */

   char file[0x1000];
   char rfile[0x1000];
   struct stat s;

   assert(c != NULL);
   assert(r_path != NULL);
   assert(d_name != NULL);
   
   /* Qualify the filename */
   qualify_full_path_file(c, file, sizeof(file), r_path, d_name);
   qualify_rel_path_file(c, rfile, sizeof(rfile), r_path, d_name);

   /* Skip if a dot-file, unless it is '..' */
   if(*d_name == '.' && !strequal(d_name, "..")) return(false);
   
   /* Stat the directory entry */
   if(stat(file, &s) != 0) return(false);
   
   /* We can list it! */
   if(!S_ISDIR(s.st_mode) && !S_ISLNK(s.st_mode)) return(false);

   /* Check file permissions */
   if(targeted_check_access(c, rfile) != ACCESS_OK) return(false);

   /* Okay to add */
   return(true);

}


bool not_is_directory(const config *c, const char *r_path, const char *d_name) {
/* Returns true if the relative path listed is NOT a directory, 
   and the user has the appropriate access rights to it. */

   char file[0x1000];
   char rfile[0x1000];
   struct stat s;

   assert(c != NULL);
   assert(r_path != NULL);
   assert(d_name != NULL);
   
   /* Qualify the filename */
   qualify_full_path_file(c, file, sizeof(file), r_path, d_name);
   qualify_rel_path_file(c, rfile, sizeof(rfile), r_path, d_name);

   /* Skip if a dot-file, unless it is '..' */
   if(*d_name == '.' && !strequal(d_name, "..")) return(false);
   
   /* Stat the directory entry */
   if(stat(file, &s) != 0) return(false);
   
   /* We can list it! */
   if(S_ISDIR(s.st_mode) || S_ISLNK(s.st_mode)) return(false);

   /* Check file permissions */
   if(targeted_check_access(c, rfile) != ACCESS_OK) return(false);

   /* Okay to add */
   return(true);

}


bintree *targeted_list_dir(const config *c, const char *r_path, listfilterfn filter) {

   int acc;
   char path[0x1000];
   bintree *tree;
   struct dirent *dir;
   DIR *d;

   assert(c != NULL);
   assert(r_path != NULL);
   assert(filter != NULL);

   /* Make sure we're allowed a directory list here */
   if(strequal(r_path, "/") && filter == is_directory) {
      /* Accept / for this special-case */
   } else {
      if(!validate_directory_name(r_path)) {
         targeted_log_error(c, "invalid directory specification");
         return(NULL);
      }
      if(targeted_index_present(c, r_path)) {
         targeted_log_error(c, "no access allowed when index.(s)html present");
         return(NULL);
      }
   } /* Sanity checks bypassed for / */
   acc = targeted_check_access(c, r_path);
   if(acc != ACCESS_OK) {
      targeted_log_error_access(c, acc);
      return(NULL);
   }
   
   /* List the directory entry names */
   qualify_full_path(c, path, sizeof(path), r_path);
   if((d = opendir(path)) == NULL) {
      targeted_log_error(c, "Failed to open directory");
      return(NULL);
   }

   /* Construct an R-B tree to store the list in */
   tree = bintree_new(bintree_string_lessthan, free);
   
   /* Read the list of files in */
   while((dir = readdir(d)) != NULL) {
      if(filter(c, r_path, dir->d_name)) {
         /* Filter accepted it */
         bintree_insert(tree, strdupl(dir->d_name));
      } /* Check with filter */
   } /* Listing files */
   closedir(d);
   
   /* Return the tree */
   return(tree);

}


typedef struct _prev_data {
   bool found;
   const char *last;
   const char *search;
} prev_data;


static bool _prev_entry(const void *_data, void *_params) {
/* Filter function for traversal function */

   const char *d_name = (const char *)_data;
   prev_data *params = (prev_data *)_params;
   
   /* Check if we should overwrite last */
   if(params->found) return(false);

   /* Check if this entry matches */
   if(strequal(d_name, params->search)) {
      /* Match; return last entry found */
      params->found = true;
      return(false);
   }
   
   /* Continue */
   params->last = d_name;
   return(true);

}


const char *targeted_prev(const config *c, const bintree *list, const char *d_name) {
/* Return the entry immediately prior to this entry, or 
   NULL if no such entry exists. */

   prev_data params;
   
   assert(c != NULL);
   assert(list != NULL);
   assert(d_name != NULL);

   params.found = false;   
   params.last = NULL;
   params.search = d_name;
   bintree_walk(list, _prev_entry, &params);
   if(params.found) return(params.last);
   return(NULL);

}


typedef struct _next_data {
   bool found;
   const char *next;
   const char *search;
} next_data;


static bool _next_entry(const void *_data, void *_params) {
/* Filter function for tree traversal */

   const char *d_name = (const char *)_data;
   next_data *params = (next_data *)_params;

   /* Did we already find a match? */
   if(params->found && params->next == NULL) {
      params->next = d_name;
      return(false);
   }

   /* Check if this entry matches */
   if(strequal(d_name, params->search)) {
      /* Match; prepare to return next entry */
      params->found = true;
   }
   return(true);
   
}


const char *targeted_next(const config *c, const bintree *list, const char *d_name) {
/* Return the entry immediately after to this entry, or 
   NULL if no such entry exists. */
   
   next_data params;
   
   assert(c != NULL);
   assert(list != NULL);
   assert(d_name != NULL);
   
   params.next = NULL;
   params.found = false;
   params.search = d_name;
   bintree_walk(list, _next_entry, &params);
   return(params.next);

}


typedef struct _child_data {
   int max;
   int found;
   const char **children;
} child_data;


static bool _child_entry(const void *_data, void *_params) {
/* Filter function for tree traversal */

   const char *d_name = (const char *)_data;
   child_data *params = (child_data *)_params;

   /* Check if this entry matches */
   if(!strequal(d_name, ".") && !strequal(d_name, "..")) {
      /* Match; prepare to return next entry */
      if(params->found < params->max) {
         params->children[params->found] = d_name;
      }
      ++params->found;
   }
   return(true);
   
}


bool targeted_children(const config *c, const char **children, int max, const bintree *list) {
/* Return the first children *directory* of this directory 
   that is not a "." or ".." directory.  The list should be 
   of the *current* directory, not the parent's tree. */
   
   child_data params;
   
   assert(c != NULL);
   assert(children != NULL);
   assert(max >= 0);
   assert(list != NULL);
   
   params.max = max;
   params.found = 0;
   params.children = children;
   bintree_walk(list, _child_entry, &params);
   return(params.found > params.max);
   
}
