/* $Header: /fridge/cvs/cgi.targeted/targeted-util.c,v 1.10 2001/12/10 23:19:05 justins Exp $ */
/*
   
   targeted-util.c            Copyright(c) 2001-1998 Justin David Smith
   justins(at)chaos2.org      http://chaos2.org/
    
   Miscellaneous utilities 
    

   This program is free software; you can redistribute it and/or modify 
   it under the terms of the GNU General Public License as published by 
   the Free Software Foundation, version 2 of the License ONLY.

   This program is distributed in the hope that it will be useful, 
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation, 
   Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/


/* Includes */
#include <targeted-util.h>
#include <targeted-log.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <errno.h>


struct timeval tv_timer;


void timer_init(void) {
/* Initialise the runtime timer */

   gettimeofday(&tv_timer, NULL);

}


double time_elapsed(void) {
/* Return the current time elapsed */

   struct timeval tv_end;
   long seconds, useconds;
   gettimeofday(&tv_end, NULL);
   seconds  = tv_end.tv_sec  - tv_timer.tv_sec;
   useconds = tv_end.tv_usec - tv_timer.tv_usec;
   while(useconds < 0) {
      --seconds;
      useconds += 1000000;
   }
   return(seconds + useconds / 1000000.0);

}


bool legal_input(const char *input) {
/* Make sure the input is non-NULL, and doesn't contain
   illegal characters (chars with special meaning in HTML) */
   
   if(input == NULL) return(false);
   while(*input != '\0') {
      /* First few are HTML escapes; % is a guard against
         any yet-unpacked CGI chars; | is special in our
         data/logfiles so we disallow it as well. */
      if(*input == '<' || *input == '>' || 
         *input == '&' || *input == '%' ||
         *input == '|') return(false);
      ++input;
   }
   return(true);
   
}


void qualify_full_path_file(const config *c, char *dest, int size, const char *r_path, const char *d_name) {
/* Qualify a full name based on path and file components */

   assert(c != NULL);
   assert(dest != NULL);
   assert(size >= 0);
   assert(r_path != NULL);
   assert(d_name != NULL);
   sbprintf(dest, size, "%s%s%s", stra(c->dir_prefix_fs), r_path, d_name);
   
}


void qualify_rel_path_file(const config *c, char *dest, int size, const char *r_path, const char *d_name) {
/* Qualify a relative name based on path and file components */

   assert(c != NULL);
   assert(dest != NULL);
   assert(size >= 0);
   assert(r_path != NULL);
   assert(d_name != NULL);
   sbprintf(dest, size, "%s%s", r_path, d_name);
   
}


void qualify_full_path(const config *c, char *dest, int size, const char *r_path) {
/* Qualify a full name based on a relative path. */

   assert(c != NULL);
   assert(dest != NULL);
   assert(size >= 0);
   assert(r_path != NULL);
   sbprintf(dest, size, "%s%s", stra(c->dir_prefix_fs), r_path);
   
}


bool compute_parent_directory(char *dest, int size, const char *r_path) {
/* Compute the name of the parent (relative) directory; return
   true if we were successful at computing it. */

   const char *p;
   const char *pterm = NULL;

   /* Sanity checks */
   assert(dest != NULL);
   assert(size > 0);
   assert(r_path != NULL);
   
   /* Check that path not already / */
   if(strequal(r_path, "/")) return(false);
   
   /* Find last matching / */
   p = r_path + strlenn(r_path) - 1;
   while(p > r_path && pterm == NULL) {
      --p;
      if(*p == '/') pterm = p + 1;
   }
   if(pterm == NULL) return(false);
   
   /* Truncate and copy */
   strcopynb(dest, r_path, pterm - r_path, size);
   return(true);
   
}


bool targeted_index_present(const config *c, const char *r_path) {
/* Check if an index file is present in the relative path given. */

   char filename[0x1000];
   const slinknode *cur;
   struct stat s;
   
   assert(c != NULL);
   assert(r_path != NULL);
   
   /* Search through the list of possible files */
   for(cur = c->index_names; cur != NULL; cur = cur->next) {
      qualify_full_path_file(c, filename, sizeof(filename), r_path, stra((const stringa *)cur->data));
      if(stat(filename, &s) == 0) return(true);
   }
   return(false);
   
}

   
bool validate_directory_name(const char *r_path) {

   const char *p = r_path;

   /* Sanity checks */
   assert(r_path != NULL);
   
   /* Validate the pathname begins with / */
   if(*r_path != '/') return(false);
   if(*(r_path + 1) == '\0') return(false);
   
   /* Validate entries in the path */
   while(*p != '\0') {
      if(*p == '/') {
         ++p;
         if(*p == '\0') return(true);
         if(*p == '/') return(false);
         if(*p == '.' && (*(p + 1) == '.' || *(p + 1) == '/')) return(false);
      }
      if((*p < 'a' || *p > 'z') &&
         (*p < 'A' || *p > 'Z') &&
         (*p < '0' || *p > '9') &&
         (*p != '-' && *p != '.' && *p != '_')) return(false);
      ++p;
   }
   
   /* Path didn't end in / */
   return(false);

}


bool validate_file_name(const char *r_path) {

   const char *p = r_path;

   /* Sanity checks */
   assert(r_path != NULL);
   
   /* Validate the pathname begins with / */
   if(*r_path != '/') return(false);
   if(*(r_path + 1) == '\0') return(false);
   
   /* Validate entries in the path */
   while(*p != '\0') {
      if(*p == '/') {
         ++p;
         if(*p == '\0') return(false);
         if(*p == '/') return(false);
         if(*p == '.' && (*(p + 1) == '.' || *(p + 1) == '/')) return(false);
      }
      if((*p < 'a' || *p > 'z') &&
         (*p < 'A' || *p > 'Z') &&
         (*p < '0' || *p > '9') &&
         (*p != '-' && *p != '.' && *p != '_')) return(false);
      ++p;
   }
   
   /* Path didn't end in / */
   return(true);

}


static bool _check_perms(uid_t uid, gid_t gid, mode_t mode) {
/* Check that permissions are okay. */

   if(S_ISDIR(mode)) {
      if(uid == geteuid()) return((mode & S_IRUSR) != 0 && (mode & S_IXUSR) != 0);
      if(gid == getegid()) return((mode & S_IRGRP) != 0 && (mode & S_IXGRP) != 0);
      return((mode & S_IROTH) != 0 && (mode & S_IXOTH) != 0);
   } else {
      if(uid == geteuid()) return((mode & S_IRUSR) != 0);
      if(gid == getegid()) return((mode & S_IRGRP) != 0);
      return((mode & S_IROTH) != 0);
   }

}


static int _check_control(const config *c, const char *ip, const char *r_path) {

   char buf[0x1000];
   char htaccess[0x1000];
   bool deny_first = true;
   bool accepted = false;
   bool denied = false;
   FILE *f;

   /* Is a control file present in this directory? */
   sbprintf(buf, sizeof(buf), "/%s", stra(c->htaccess_name));
   qualify_full_path_file(c, htaccess, sizeof(htaccess), r_path, buf);
   if((f = fopen(htaccess, "r")) != NULL) {
      /* Scan the control file for our IP prefix */
      while(fgets(buf, sizeof(buf), f) != NULL) {
         trim(buf);
         strlower(buf);
         if(strequaln(buf, "order", 5)) {
            /* Order directive */
            const char *pallow = rkstrpat(NULL, buf, "allow");
            const char *pdeny = rkstrpat(NULL, buf, "deny");
            deny_first = (pdeny < pallow);
         } else if(strequaln(buf, "allow", 5)) {
            /* Allow directive */
            char *pfrom = rkstrpat(NULL, buf, "from");
            if(pfrom != NULL) {
               pfrom += 4;
               trim(pfrom);
               if(strequal(pfrom, "all")) accepted = true;
               else if(strequaln(pfrom, ip, strlenn(pfrom))) accepted = true;
            }
         } else if(strequaln(buf, "deny", 4)) {
            /* Deny directive */
            char *pfrom = rkstrpat(NULL, buf, "from");
            if(pfrom != NULL) {
               pfrom += 4;
               trim(pfrom);
               if(strequal(pfrom, "all")) denied = true;
               else if(strequaln(pfrom, ip, strlenn(pfrom))) denied = true;
            }
         } /* Other (unknown) directives */
      } /* Scanning control file */
      fclose(f);
      
      /* Check to see if we were allowed/denied here */
      if(accepted) return(ACCESS_OK);
      if(denied) return(HTACCESS_DENIES);
   } /* All over */
   
   /* Try next directory up */
   if(compute_parent_directory(buf, sizeof(buf), r_path)) {
      return(_check_control(c, ip, buf));
   } else {
      /* Succeed by default */
      return(ACCESS_OK);
   }

}


int targeted_check_access(const config *c, const char *r_path) {

   char path[0x1000];
   const char *ip;
   struct stat s;
   
   assert(c != NULL);
   assert(r_path != NULL);
   
   /* Get file status */
   qualify_full_path(c, path, sizeof(path), r_path);
   if(stat(path, &s) != 0) {
      if(errno == EACCES) return(PERM_DENIES);
      if(errno == ENOENT) return(FILE_NOT_FOUND);
      return(UNKNOWN_DENIES);
   }

   /* Check fs permissions */
   if(!_check_perms(s.st_uid, s.st_gid, s.st_mode)) {
      return(PERM_DENIES);
   }
   
   /* Check control file permissions */
   ip = stra(c->env.remote_addr);
   return(_check_control(c, ip, r_path));

}


stringa *rewrite_directory_name(const char *r_path) {
/* If directory not ending in "/", append a slash to it.
   Make sure to strfreea() the resulting string... */

   stringa *result = strnewa(r_path);
   const char *p = r_path;
   
   if(r_path == NULL || *r_path == '\0') return(result);
   while(*p != '\0') ++p;
   --p;
   
   if(*p == '/') return(result);
   strconcata(result, "/");
   return(result);

}


void print_file_size(FILE *f, int size) {
/* Prints a file size with appropriate units */

   if(size > 1024 * 1024) {
      fprintf(f, "%dMB", size / (1024 * 1024));
   } else if(size > 1024) {
      fprintf(f, "%dKB", size / 1024);
   } else {
      fprintf(f, "%dB", size);
   }

}


bool targeted_admin_ip(const config *c) {
/* Returns true if the remote host is a valid administrative IP */

   ip_number ip;
   const slinknode *cur;

   if(!ip_number_init(&ip, stra(c->env.remote_addr))) return(false);

   for(cur = c->admin_ip; cur != NULL; cur = cur->next) {
      if(ip_number_match((const ip_number *)cur->data, &ip)) return(true);
   }
   return(false);
   
}
